/* $File: //ASP/Dev/SBS/4_Controls/4_3_Network_Infrastructure/4_3_1_Comms_Common_Services/sw/device_drivers/MKS937A/MKS937ASup/devMKS937A.c $
 * $Revision: #17 $
 * $DateTime: 2008/02/05 17:12:36 $
 * Last checked in by: $Author: starritt $
 *
 * Description
 * MKS937A device support.
 *
 * Formatted using:
 *    indent -kr -pcs -i3 -cli3 -nut
 *
 * Copyright  2005  Australian Synchrotron
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * Licence as published by the Free Software Foundation; either
 * version 2.1 of the Licence, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public Licence for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * Licence along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Contact details:
 * andrew.starritt@synchrotron.vic.gov.au
 * 800 Blackburn Road, Clayton, Victoria 3168, Australia.
 *
 * -----------------------------------------------------------------------------
 *
 * Modifications for using the device support for the ASP SXR-Beamline:
 * - read actual status of relays
 * - write support (for adjusting the relay setpoints)
 *
 * Date: 2006/11/07
 * Jens Rekow
 * jens.rekow@enz-berlin.de
 * ENZ Ingenieurbuero
 * -----------------------------------------------------------------------------
 */

#include <alarm.h>
#include <devCommonGpib.h>
#include <epicsStdio.h>
#include <recGbl.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>


/******************************************************************************
 *
 * The following define statements are used to declare the names to be used
 * for the dset tables.
 *
 * A DSET_AI entry must be declared here and referenced in an application
 * database description file even if the device provides no AI records.
 *
 ******************************************************************************
 */

#define DSET_AI     devAiMKS937A
#define DSET_MBBI   devMbbiMKS937A
#define DSET_SI     devSiMKS937A
#define DSET_LI     devLiMKS937A
#define DSET_AO     devAoMKS937A

#include <devGpib.h>            /* must be included after DSET defines */


/* ENZ changed both of these timeout values to 0.5,
 * but we left as original values (for now)
*/
#define TIMEOUT     5.0         /* I/O must complete within this time  */
#define TIMEWINDOW  2.0         /* Wait this long after device timeout */


/******************************************************************************
 */
static char *Controller_Status[] = {
   "",                          /* place holder value - used for Unknown */
   "",                          /* place holder value - used for Invalid */
   "",                          /* place holder value - used for Okay    */
   "LO",                        /* 3 - used for LO<E-dd format as well.  */
   "WAIT",                      /* 4 */
   "HV_OFF!",                   /* 5 */
   "NOGAUGE!",                  /* 6 */
   "MISCONN!",                  /* 7 */
   "PROTECT!",                  /* 8 */
   "CONTROL!",                  /* 9 */
   "LowEmis!",                  /* 10 */
   "NEGATIV!",                  /* 11 */
   NULL
};

#define INVALID_STATUS      1
#define OKAY_STATUS         2
#define LOW_STATUS          3
#define HIGH_STATUS        12
#define ATMOSPHERE_STATUS  13


static char *Comms_Status[] = { "",     /* place holder value - used for Unknown */
   "COMLOCK",
   "COMFREE",
   NULL
};

#define UNKNOWN_COMMS  0


static char *Cal_Status[] = { "",       /* place holder value - used for Unknown */
   "CALLOCK",
   "CALFREE",
   NULL
};

#define UNKNOWN_CAL    0


static char *Front_Status[] = { "",     /* place holder value - used for Unknown */
   "FP LOCK",
   "FP FREE",
   NULL
};

#define UNKNOWN_FRONT  0


/******************************************************************************
 * Result values for the convert reply functions.
 */
#define OK    0
#define NOTOK 1

/******************************************************************************
 * Sorts of Pressure values read from gauge
 * Supplied as parameter P2 to Convert_Pressure_Reply procedure
 */
#define GAUGE_PRESSURE         1
#define RELAY_SET_POINT        2
#define PROTECTION_SET_POINT   3
#define CONTROL_SET_POINT      4
#define SENSITIVITY_VALUE      5


/******************************************************************************
 * Declare functions for gpibCmds.
 ******************************************************************************
 */
static int Convert_Status_Reply_To_String
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3);

static int Convert_Status_Reply_To_MBBI
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3);

static int Convert_Comms_Reply_To_MBBI
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3);

static int Convert_Cal_Reply_To_MBBI
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3);

static int Convert_Front_Reply_To_MBBI
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3);

static int Convert_Pressure_Reply
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3);

static int Convert_BitField_Reply_To_Long
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3);


/******************************************************************************
 * Array of structures that define the MKS 937A Controller commands.
 *
 * We essentially use just one command, the Pn command, to interrogate the
 * pressure/status of each channel.  This is replicated five times for
 * each of the potential channels (although we expect to be using two or
 * three channels) by three for each record type (SI, MBBI and AI).
 *
 ******************************************************************************
 */

static struct gpibCmd gpibCmds[] = {

    /****** Status Reply - SI ******/
   /* Param 0 */
   {&DSET_SI, GPIBREAD, IB_Q_LOW, "$0P1\r", NULL, 0, 20,
    Convert_Status_Reply_To_String, 1, 0, NULL, NULL, "\r"},

   /* Param 1 */
   {&DSET_SI, GPIBREAD, IB_Q_LOW, "$0P2\r", NULL, 0, 20,
    Convert_Status_Reply_To_String, 2, 0, NULL, NULL, "\r"},

   /* Param 2 */
   {&DSET_SI, GPIBREAD, IB_Q_LOW, "$0P3\r", NULL, 0, 20,
    Convert_Status_Reply_To_String, 3, 0, NULL, NULL, "\r"},

   /* Param 3 */
   {&DSET_SI, GPIBREAD, IB_Q_LOW, "$0P4\r", NULL, 0, 20,
    Convert_Status_Reply_To_String, 4, 0, NULL, NULL, "\r"},

   /* Param 4 */
   {&DSET_SI, GPIBREAD, IB_Q_LOW, "$0P5\r", NULL, 0, 20,
    Convert_Status_Reply_To_String, 5, 0, NULL, NULL, "\r"},


    /****** Status Reply - MBBI ******/
   /* Param 5 */
   {&DSET_MBBI, GPIBREAD, IB_Q_LOW, "$0P1\r", NULL, 0, 20,
    Convert_Status_Reply_To_MBBI, 1, 0, NULL, NULL, "\r"},

   /* Param 6 */
   {&DSET_MBBI, GPIBREAD, IB_Q_LOW, "$0P2\r", NULL, 0, 20,
    Convert_Status_Reply_To_MBBI, 2, 0, NULL, NULL, "\r"},

   /* Param 7 */
   {&DSET_MBBI, GPIBREAD, IB_Q_LOW, "$0P3\r", NULL, 0, 20,
    Convert_Status_Reply_To_MBBI, 3, 0, NULL, NULL, "\r"},

   /* Param 8 */
   {&DSET_MBBI, GPIBREAD, IB_Q_LOW, "$0P4\r", NULL, 0, 20,
    Convert_Status_Reply_To_MBBI, 4, 0, NULL, NULL, "\r"},

   /* Param 9 */
   {&DSET_MBBI, GPIBREAD, IB_Q_LOW, "$0P5\r", NULL, 0, 20,
    Convert_Status_Reply_To_MBBI, 5, 0, NULL, NULL, "\r"},


    /****** Pressure Reply - AI ******/
   /* Param 10 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0P1\r", NULL, 0, 20,
    Convert_Pressure_Reply, 1, GAUGE_PRESSURE, NULL, NULL, "\r"},

   /* Param 11 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0P2\r", NULL, 0, 20,
    Convert_Pressure_Reply, 2, GAUGE_PRESSURE, NULL, NULL, "\r"},

   /* Param 12 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0P3\r", NULL, 0, 20,
    Convert_Pressure_Reply, 3, GAUGE_PRESSURE, NULL, NULL, "\r"},

   /* Param 13 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0P4\r", NULL, 0, 20,
    Convert_Pressure_Reply, 4, GAUGE_PRESSURE, NULL, NULL, "\r"},

   /* Param 14 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0P5\r", NULL, 0, 20,
    Convert_Pressure_Reply, 5, GAUGE_PRESSURE, NULL, NULL, "\r"},


    /****** Relay Set Point Reply - AI ******/
   /* Param 15 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0RLY1\r", NULL, 0, 20,
    Convert_Pressure_Reply, 1, RELAY_SET_POINT, NULL, NULL, "\r"},

   /* Param 16 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0RLY2\r", NULL, 0, 20,
    Convert_Pressure_Reply, 2, RELAY_SET_POINT, NULL, NULL, "\r"},

   /* Param 17 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0RLY3\r", NULL, 0, 20,
    Convert_Pressure_Reply, 3, RELAY_SET_POINT, NULL, NULL, "\r"},

   /* Param 18 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0RLY4\r", NULL, 0, 20,
    Convert_Pressure_Reply, 4, RELAY_SET_POINT, NULL, NULL, "\r"},

   /* Param 19 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0RLY5\r", NULL, 0, 20,
    Convert_Pressure_Reply, 5, RELAY_SET_POINT, NULL, NULL, "\r"},


    /****** Protection Set Point Reply - AI ******/
   /* Param 20 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0PRO1\r", NULL, 0, 20,
    Convert_Pressure_Reply, 1, PROTECTION_SET_POINT, NULL, NULL, "\r"},

   /* Param 21 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0PRO2\r", NULL, 0, 20,
    Convert_Pressure_Reply, 2, PROTECTION_SET_POINT, NULL, NULL, "\r"},

   /* Param 22 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0PRO3\r", NULL, 0, 20,
    Convert_Pressure_Reply, 3, PROTECTION_SET_POINT, NULL, NULL, "\r"},

   /* Param 23 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0PRO4\r", NULL, 0, 20,
    Convert_Pressure_Reply, 4, PROTECTION_SET_POINT, NULL, NULL, "\r"},

   /* Param 24 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0PRO5\r", NULL, 0, 20,
    Convert_Pressure_Reply, 5, PROTECTION_SET_POINT, NULL, NULL, "\r"},


    /****** Monitor Control Set Point Reply - AI ******/
   /* Param 25 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0CTL1\r", NULL, 0, 20,
    Convert_Pressure_Reply, 1, CONTROL_SET_POINT, NULL, NULL, "\r"},

   /* Param 26 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0CTL2\r", NULL, 0, 20,
    Convert_Pressure_Reply, 2, CONTROL_SET_POINT, NULL, NULL, "\r"},

   /* Param 27 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0CTL3\r", NULL, 0, 20,
    Convert_Pressure_Reply, 3, CONTROL_SET_POINT, NULL, NULL, "\r"},

   /* Param 28 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0CTL4\r", NULL, 0, 20,
    Convert_Pressure_Reply, 4, CONTROL_SET_POINT, NULL, NULL, "\r"},

   /* Param 29 */
   {&DSET_AI, GPIBCVTIO, IB_Q_LOW, "$0CTL5\r", NULL, 0, 20,
    Convert_Pressure_Reply, 5, CONTROL_SET_POINT, NULL, NULL, "\r"},


   /* Specials */
    /****** COM Reply - MBBI ******/
   /* Param 30 */
   {&DSET_MBBI, GPIBREAD, IB_Q_LOW, "$0COM\r", NULL, 0, 20,
    Convert_Comms_Reply_To_MBBI, 0, 0, NULL, NULL, "\r"},

    /****** CAL Reply - MBBI ******/
   /* Param 31 */
   {&DSET_MBBI, GPIBREAD, IB_Q_LOW, "$0CAL\r", NULL, 0, 20,
    Convert_Cal_Reply_To_MBBI, 0, 0, NULL, NULL, "\r"},

    /****** FRONT Reply - MBBI ******/
   /* Param 32 */
   {&DSET_MBBI, GPIBREAD, IB_Q_LOW, "$0FRONT\r", NULL, 0, 20,
    Convert_Front_Reply_To_MBBI, 0, 0, NULL, NULL, "\r"},

   /* Setpoint Relays */
   /****** Status of Relays - LI ******/
   /* Param 33 */
   {&DSET_LI, GPIBREAD, IB_Q_LOW, "$0RELAYS\r", NULL, 0, 20,
    Convert_BitField_Reply_To_Long, 0, 0, NULL, NULL, "\r"},

   /* param 34 - spare - place holder */
   FILL,

    /****** Set Control Set Point Reply - A0 ******/
   /* Param 35 */
   {&DSET_AO, GPIBWRITE, IB_Q_LOW, NULL, "$0RLY1=%1.1E\r", 20, 20,
    NULL, 0, 0, NULL, NULL, "\r"},

   /* Param 36 */
   {&DSET_AO, GPIBWRITE, IB_Q_LOW, NULL, "$0RLY2=%1.1E\r", 20, 20,
    NULL, 0, 0, NULL, NULL, "\r"},

   /* Param 37 */
   {&DSET_AO, GPIBWRITE, IB_Q_LOW, NULL, "$0RLY3=%1.1E\r", 20, 20,
    NULL, 0, 0, NULL, NULL, "\r"},

   /* Param 38 */
   {&DSET_AO, GPIBWRITE, IB_Q_LOW, NULL, "$0RLY4=%1.1E\r", 20, 20,
    NULL, 0, 0, NULL, NULL, "\r"},

   /* Param 39 */
   {&DSET_AO, GPIBWRITE, IB_Q_LOW, NULL, "$0RLY5=%1.1E\r", 20, 20,
    NULL, 0, 0, NULL, NULL, "\r"}

};

/* The following is the number of elements in the command array above.  */

#define NUMPARAMS sizeof(gpibCmds)/sizeof(struct gpibCmd)


/******************************************************************************
 * Procedure and function bodies
 ******************************************************************************
 */

/******************************************************************************
 * Initialize device support parameters
 */
static long init_ai (int parm)
{
   printf ("devMKS937A.init_ai (%d) invoked \n", parm);

   if (parm == 0) {
      devSupParms.name = "devMKS937A";
      devSupParms.gpibCmds = gpibCmds;
      devSupParms.numparams = NUMPARAMS;
      devSupParms.timeout = TIMEOUT;
      devSupParms.timeWindow = TIMEWINDOW;

      /* Set >= 0 if device responds to writes.
       * We ARE using write commands now.
       */
      devSupParms.respond2Writes = 0;
   }
   return (0);

}                               /* init_ai */


/******************************************************************************
 * Searches the set of strings in List for the specified string Item, and
 * returns the index if found, otherwise returns -1.
 */
static int Index_Of_Item (char *Item, char *List[])
{

   int Result;
   int Index;

   Result = -1;

   for (Index = 0; List[Index] != NULL; Index++) {

      if (strcmp (Item, List[Index]) == 0) {
         /* found it */
         Result = Index;
         break;
      }

   }
   return Result;

}                               /* Index_Of_Item  */


/******************************************************************************
 */
static void Process_Gauge_Reply (char *Reply, double *pressure, int *rval)
{

   int p;

   /* Attempt to read input as a floating point number
    *
    * Expect to read one field.
    */
   if (sscanf (Reply, "%le", pressure) == 1) {

      /* We successfully read one number. Set rval to OKAY.
       */
      *rval = OKAY_STATUS;


   } else if (sscanf (Reply, "LO<E%d", &p) == 1) {

      /* We successfully read a low limit value.
       */
      *rval = LOW_STATUS;

      /* Special case for Gauge Pressure LO<E-xx status.
       */
      *pressure = pow (10.0, p);


   } else if (sscanf (Reply, "HI>E%d", &p) == 1) {

      /* We successfully read a high limit value.
       */
      *rval = HIGH_STATUS;

      /* Special case for Gauge Pressure HI>E-xx status.
       */
      *pressure = pow (10.0, p);


   } else if (sscanf (Reply, "AA_E%d", &p) == 1) {

      /* We successfully read an at atmosphere limit value.
       */
      *rval = ATMOSPHERE_STATUS;

      /* Special case for Gauge Pressure AA_E-xx status.
       */
      *pressure = pow (10.0, p);


   } else {

      /* Compare status string with known/expected status.
       */
      *rval = Index_Of_Item (Reply, Controller_Status);

      if (*rval == -1) {
         *rval = INVALID_STATUS;
      }

      /* No explicit pressure value within the reply.
       */
      *pressure = -1.0;

   }

}                               /* Process_Gauge_Reply */


/******************************************************************************
 */
static int Convert_Status_Reply_To_MBBI
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3) {

   struct mbbiRecord *pmbbi = ((struct mbbiRecord *) (pdpvt->precord));
   double pressure;
   int rval;

   /* printf ("devMKS937A.Convert_Status_Reply_To_MBBI invoked \n");  */

   Process_Gauge_Reply (pdpvt->msg, &pressure, &rval);

   /* Update the record with rval.
    */
   pmbbi->rval = rval;
   pmbbi->udf = 0;

   return OK;

}                               /* Convert_Status_Reply_To_MBBI */


/******************************************************************************
 */
static int Convert_Pressure_Reply
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3) {

   struct aiRecord *pai = ((struct aiRecord *) (pdpvt->precord));

   asynOctet *pasynOctet = pdpvt->pasynOctet;
   void *asynOctetPvt = pdpvt->asynOctetPvt;
   gpibCmd *pgpibCmd = gpibCmdGet (pdpvt);
   asynStatus status;

   size_t ntrans;
   char response_buffer[20];
   int numFields;
   char units[20];

   int result;
   int rval;
   double pressure;
   int units_required;
   int size;

   /* printf ("devMKS937A.Convert_Pressure_Reply invoked \n"); */

   /* Assume it's not going to work - set defaults.
    */
   result = NOTOK;
   pressure = 0.0;

   /* Commence IO
    * Set the End of String character to that specified in the command.
    */
   pdevSupportGpib->setEos (pdpvt, pgpibCmd);


   /* Always request unist from the device.
    */
   units_required = 1;

   if (units_required == 1) {

      status = pasynOctet->write
          (asynOctetPvt, pdpvt->pasynUser, "$0UNIT\r",
           sizeof "$0UNIT\r" - 1, &ntrans);

      if (status == asynSuccess) {

         status = pasynOctet->read
             (asynOctetPvt, pdpvt->pasynUser, response_buffer,
              sizeof response_buffer - 1, &ntrans, NULL);

         if (status == asynSuccess) {

            response_buffer[ntrans] = '\0';

            /* Scanf essentially does a trim
             */
            numFields = sscanf (response_buffer, "%s", units);

            if (numFields == 1) {

               /* Convert pressure units into standard form */

               if (strcmp (units, "mbar") == 0) {
                  strcpy (pai->egu, "mbar");

               } else if (strcmp (units, "Torr") == 0) {
                  strcpy (pai->egu, "Torr");

               } else if (strcmp (units, "Pascal") == 0) {
                  strcpy (pai->egu, "Pa");

               } else if (strcmp (units, "micron") == 0) {
                  strcpy (pai->egu, "um");

               } else {
                  strcpy (pai->egu, "???");

               }

            } else {

               /* Read of units failed */
               strcpy (pai->egu, "???");
            }

         }
      }
   }


   /* Now write the standard command
    */
   size = strlen (pgpibCmd->cmd);

   status = pasynOctet->write
       (asynOctetPvt, pdpvt->pasynUser, pgpibCmd->cmd, size, &ntrans);

/*  printf ("\n sent \n%s\n size %d \n ntrans %d \n", pgpibCmd->cmd, size, ntrans);
 */

   if (status == asynSuccess) {

      status = pasynOctet->read
          (asynOctetPvt, pdpvt->pasynUser, response_buffer,
           sizeof response_buffer - 1, &ntrans, NULL);

      if (status == asynSuccess) {

         /* Attempt to extract pressure and/or status.
          */
         Process_Gauge_Reply (response_buffer, &pressure, &rval);

         if (pressure >= 0.0) {

            /* We successfully read a pressure.
               This is in the EGU units specified by the device.
             */
            result = OK;
         }

      }

   }

   /* IO complete - restore the End of String characters.
    */
   pdevSupportGpib->restoreEos (pdpvt, pgpibCmd);


   /* If all went okay, update the record with the read pressure
    * value (converted to EGU units), otherwise set the record
    * undefined and set the pressure to zero.
    */
   if (result == OK) {

      pai->val = pressure;
      pai->nsev = NO_ALARM;
      pai->nsta = NO_ALARM;
      pai->udf = 0;

   } else {

      pai->val = 0.0;
      recGblSetSevr (pai, READ_ALARM, INVALID_ALARM);
      pai->udf = 1;

   }

   return result;

}                               /* Convert_Pressure_Reply */


/******************************************************************************
 */
static int Convert_Status_Reply_To_String
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3) {

   struct stringinRecord *psi =
       ((struct stringinRecord *) (pdpvt->precord));
   int Result;
   int Status;
   double pressure;
   int Length;

   /* printf ("devMKS937A.Convert_Status_Reply_To_String invoked \n");  */

   /* Attempt to read input as a floating point number.
    */
   Status = sscanf (pdpvt->msg, "%le", &pressure);

   /* Expect to read one field.
    */
   if (Status == 1) {

      /* We successfully read one number. Set val to OKAY.
       */
      strcpy (psi->val, "OKAY");
      Result = OK;

   } else {

      /* Not a number - just copy status string.
       *
       * Will returned string fit into the record?
       */
      Length = pdpvt->msgInputLen;
      if (Length < 40) {

         /* Yes - copy.
          */
         strcpy (psi->val, pdpvt->msg);

         /* Remove trailing ! if it exists.
          */
         if ((Length > 0) && (psi->val[Length - 1] == '!')) {
            psi->val[Length - 1] = 0;
         }
         Result = OK;

      } else {

         epicsSnprintf (pdpvt->pasynUser->errorMessage,
                        pdpvt->pasynUser->errorMessageSize,
                        "Invalid Reply - Too long");
         Result = NOTOK;
      }
   }

   return Result;

}                               /* Convert_Status_Reply_To_String */


/******************************************************************************
 */
static int Convert_Comms_Reply_To_MBBI
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3) {

   struct mbbiRecord *pmbbi = ((struct mbbiRecord *) (pdpvt->precord));
   int rval;

   /* printf ("devMKS937A.Convert_Comms_Reply_To_MBBI invoked \n");  */

   /* Compare status string with known/expected status.
    */
   rval = Index_Of_Item (pdpvt->msg, Comms_Status);

   if (rval == -1) {
      rval = UNKNOWN_COMMS;
   }

   /* Update the record with rval.
    */
   pmbbi->rval = rval;
   pmbbi->udf = 0;

   return OK;

}                               /* Convert_Comms_Reply_To_MBBI */


/******************************************************************************
 */
static int Convert_Cal_Reply_To_MBBI
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3) {

   struct mbbiRecord *pmbbi = ((struct mbbiRecord *) (pdpvt->precord));
   int rval;

   /* printf ("devMKS937A.Convert_Cal_Reply_To_MBBI invoked \n");  */

   /* Compare status string with known/expected status.
    */
   rval = Index_Of_Item (pdpvt->msg, Cal_Status);

   if (rval == -1) {
      rval = UNKNOWN_CAL;
   }

   /* Update the record with rval.
    */
   pmbbi->rval = rval;
   pmbbi->udf = 0;

   return OK;

}                               /* Convert_Cal_Reply_To_MBBI */


/******************************************************************************
 */
static int Convert_Front_Reply_To_MBBI
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3) {

   struct mbbiRecord *pmbbi = ((struct mbbiRecord *) (pdpvt->precord));
   int rval;

   /* printf ("devMKS937A.Convert_Front_Reply_To_MBBI invoked \n");  */

   /* Compare status string with known/expected status.
    */
   rval = Index_Of_Item (pdpvt->msg, Front_Status);

   if (rval == -1) {
      rval = UNKNOWN_FRONT;
   }

   /* Update the record with rval.
    */
   pmbbi->rval = rval;
   pmbbi->udf = 0;

   return OK;

}                               /* Convert_Front_Reply_To_MBBI */


/******************************************************************************
 */
static int Convert_BitField_Reply_To_Long
    (struct gpibDpvt *pdpvt, int P1, int P2, char **P3) {

   struct longinRecord *plongin =
       ((struct longinRecord *) (pdpvt->precord));
   int rval = 0;
   int i;

   printf ("msg: %s, len: %i\n", pdpvt->msg, pdpvt->msgInputLen);

   /* while checking the msgInputLength we consider that the terminator has already been removed! */
   if (pdpvt->msgInputLen == 8) {
      pdpvt->msg[1] = 0;

      for (i = 0; i < 5; i++) {
         memmove (pdpvt->msg + 0, (pdpvt->msg + (3 + i)), 1);
         rval = rval + (atoi (pdpvt->msg) * pow (2, i));
      }

      /* Update the record with val. */
      plongin->val = rval;
      plongin->udf = 0;

      post_event (10);

      return OK;
   } else {
      epicsSnprintf (pdpvt->pasynUser->errorMessage,
                     pdpvt->pasynUser->errorMessageSize,
                     "Invalid reply by MKS 937A!\n");
      return NOTOK;
   }
}                               /* Convert_Bitfield_Reply_To_Long */

/* end */
